from netgen.csg import unit_cube
import netgen.meshing
import time

# ngsolve-imports
from ngsolve import *
## PYTHON IS SLOW
#from ngsolve.krylovspace import CGSolver
import ngs_petsc as petsc

# initialize MPI
comm = mpi_world
rank = comm.rank

# build H1-FESpace as usual
import sys
if len(sys.argv) > 1:
    order = int(sys.argv[1])
    nx = int(sys.argv[2])

do_vtk = False
domesh = False
file1 = "mesh"+str(nx)+".vol"
if rank==0 and domesh:
    # master-proc generates mesh
    mesh = unit_cube.GenerateMesh(maxh=1/nx)
    # and saves it to file
    mesh.Save(file1)

# wait for master to be done meshing
comm.Barrier()

# now load mesh from file
ngmesh = netgen.meshing.Mesh(dim=3, comm=comm)
ngmesh.Load(file1)

#refine once?
# ngmesh.Refine()

mesh = Mesh(ngmesh)

V = HDiv(mesh, order=order, discontinuous=True)
Q = L2(mesh, order=order-1+(order==0))
M = FacetFESpace(mesh, order=order, dirichlet=[1,2,3,4])
fes = FESpace([V, Q, M])
(u, p, phat), (v, q, qhat) = fes.TnT()

# RHS does not change either!
f = LinearForm (fes)
f += SymbolicLFI(32 * (y*(1-y)+x*(1-x)) * q)
f.Assemble()

# neither does the BLF!
a = BilinearForm (fes, symmetric=False, condense=True)
n = specialcf.normal(mesh.dim)
a += (-u*v+p*div(v)+div(u)*q)*dx
a += (-u*n*qhat-v*n*phat)*dx(element_boundary=True)

# BDDC + BoomerAMG for coarse matrix
jac = Preconditioner(a,"local")
#c = Preconditioner(a,"bddc", usehypre=True)
#c = Preconditioner(a,"bddc", inverse="mumps")

a.Assemble()

### PETSC solver
# NGS matrix to PETSC matrix
#
# get uhat bit array (h.o. parts)
phatBits = BitArray(fes.ndof)
phatBits[:] = 0
udof, pdof = fes.components[0].ndof, fes.components[1].ndof
phatBits[udof+pdof:] = fes.components[2].FreeDofs(True)
# remove h.o. parts
phatBits[udof+pdof+mesh.nfacet:] = 0


a1 = petsc.PETScMatrix(a.mat, phatBits)
# hypre preconditioner
#pc1 = petsc.PETSc2NGsPrecond(a1, "pc", petsc_options={"pc_type" : "lu", 
#  "pc_factor_mat_solver_type": "mumps"})
#  "pc_factor_mat_solver_type": "superlu_dist"})
pc1 = petsc.PETSc2NGsPrecond(a1, "pc", petsc_options={"pc_type" : "hypre"})
if order==0: # hypre precond
  pre = pc1
else:
  # auxillary space preconditioner
  pre = pc1 + jac.mat 
#pre = c.mat

# solve the equation
u = GridFunction (fes)
# use CG-solver with preconditioner c

for i in range(2):
     #print("Rank ", rank, ": solve started: ", datetime.now().strftime("%H:%M:%S") )
     t0 = time.time()
     f.vec.data += a.harmonic_extension_trans * f.vec
     #comm.Barrier()
     t1 = time.time()
     inv = CGSolver(a.mat, pre)
     u.vec.data = inv * f.vec  
     #comm.Barrier()
     t2 = time.time()
     u.vec.data += a.harmonic_extension * u.vec
     #comm.Barrier()
     t3 = time.time()
     u.vec.data += a.inner_solve * f.vec
     #comm.Barrier()
     t4 = time.time()
     
     #print("Rank ", rank, ": solve ended: ", datetime.now().strftime("%H:%M:%S") )
     #exact solution
     exact = 16*x*(1-x)*y*(1-y)
     
     # error 
     error = Integrate ( (u.components[1]-exact)*(u.components[1]-exact) , mesh)
     
     if rank==1:
       print("Rank %3d ncells %4d Elapse: %.2e, %.2e, %.2e, %.2e its: %3d "%(rank,
         mesh.ne, t1-t0, t2-t1, t3-t2, t4-t3, inv.GetSteps()))
